// bmp2nes.cpp

#include <stdio.h>
#include <string.h>

#define  MODE_BG2  0x0001
#define  MODE_BG2z 0x0002
#define  MODE_SP2  0x0004
#define  FILE_IN   0x0100
#define  FILE_OUT  0x0200

int  ParseArg(int argc, char *argv[], int *mode, char *inname, char *outname);
void ConvertBG2(FILE *ifp, FILE *ofp);
void ConvertBG2z(FILE *ifp, FILE *ofp);
void ConvertSP2(FILE *ifp, FILE *ofp);

int main(int argc, char *argv[]) {
	FILE *ifp = NULL;
	FILE *ofp = NULL;
	char inname[256];
	char outname[256];
	int mode;

	// 
	ParseArg(argc, argv, &mode, inname, outname);

	// t@CJ
	if ((mode & (FILE_IN|FILE_OUT)) != (FILE_IN|FILE_OUT)) {
		goto bye;
	}

	ifp = fopen(inname, "rb");
	if (ifp == NULL) {
		printf("%s ̃t@CI[vɎs܂B\n", inname);
		goto bye;
	}

	ofp = fopen(outname, "wb");
	if (ofp == NULL) {
		printf("%s ̃t@CI[vɎs܂B\n", inname);
		goto bye;
	}

	if (mode & MODE_BG2) {
		ConvertBG2(ifp, ofp);
	} else if (mode & MODE_BG2z) {
		ConvertBG2z(ifp, ofp);
	} else if (mode & MODE_SP2) {
		ConvertSP2(ifp, ofp);
	}

bye:
	if (ifp != NULL) fclose(ifp);
	if (ofp != NULL) fclose(ofp);
}

int ParseArg(int argc, char *argv[], int *mode, char *inname, char *outname) {
	int phase = 0;
	int i;

	*mode = 0;
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-bg2") == 0) {
			*mode |= MODE_BG2;
		} else if (strcmp(argv[i], "-bg2z") == 0) {
			*mode |= MODE_BG2z;
		} else if (strcmp(argv[i], "-sp2") == 0) {
			*mode |= MODE_SP2;
		} else {
			if (phase == 0) {
				strcpy(inname, argv[i]);
				*mode |= FILE_IN;
			} else if (phase == 1) {
				strcpy(outname, argv[i]);
				*mode |= FILE_OUT;
			}
			phase++;
		}
	}
}

// 8x8TCYPʂ𐅕@ɕϊ
void ConvertBG2(FILE *ifp, FILE *ofp) {
	char  head[64];
	char *src = NULL;
	char *dst = NULL;
	char *p1, *p2, *p3;
	int   offset;
	int   width;
	int   height;
	int   pitch;
	int   data;
	int   x, y, w, h;

	// wb_[̓ǂݍ
	fread(head, 54, sizeof(char), ifp);
	width   = (head[18] & 255);
	width  |= (head[19] & 255) <<  8;
	width  |= (head[20] & 255) << 16;
	width  |= (head[21] & 255) << 24;
	height  = (head[22] & 255);
	height |= (head[23] & 255) <<  8;
	height |= (head[24] & 255) << 16;
	height |= (head[25] & 255) << 24;
	pitch   = (width+3) & -4;

	// obt@̊m
	src = new char[pitch * 8];
	dst = new char[(pitch/8)*16];

	// sNZϊ
	offset = 54+1024;
	offset += pitch * height;
	for (h = 0; h < height; h += 8) {
		offset -= pitch * 8;
		fseek(ifp, offset, SEEK_SET);
		fread(src, pitch*8, sizeof(char), ifp);
		p1 = src + pitch*8;
		p3 = dst;
		for (w = 0; w < width; w += 8) {
			p2 = p1;
			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p1 += 8;
			p3 += 8;
		}
		fwrite(dst, (pitch/8)*16, sizeof(char), ofp);
	}

	// obt@̊J
	delete src;
	delete dst;
}

// 16x16TCYPʂ𐅕@ɕϊ
void ConvertBG2z(FILE *ifp, FILE *ofp) {
	char  head[64];
	char *src = NULL;
	char *dst = NULL;
	char *p1, *p2, *p3;
	int   offset;
	int   width;
	int   height;
	int   pitch;
	int   data;
	int   x, y, w, h;

	// wb_[̓ǂݍ
	fread(head, 54, sizeof(char), ifp);
	width   = (head[18] & 255);
	width  |= (head[19] & 255) <<  8;
	width  |= (head[20] & 255) << 16;
	width  |= (head[21] & 255) << 24;
	height  = (head[22] & 255);
	height |= (head[23] & 255) <<  8;
	height |= (head[24] & 255) << 16;
	height |= (head[25] & 255) << 24;
	pitch   = (width+3) & -4;

	// obt@̊m
	src = new char[pitch * 16];
	dst = new char[(pitch/8)*32];

	// sNZϊ
	offset = 54+1024;
	offset += pitch * height;
	for (h = 0; h < height; h += 16) {
		offset -= pitch * 16;
		fseek(ifp, offset, SEEK_SET);
		fread(src, pitch*16, sizeof(char), ifp);
		p1 = src + pitch*16;
		p3 = dst;
		for (w = 0; w < width; w += 16) {
			p2 = p1;
			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p3 += 8;

			p2 = p1 + 8;
			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p3 += 8;

			p2 = p1 - pitch * 8;
			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p3 += 8;

			p2 = p1 - pitch * 8 + 8;
			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p3 += 8;

			p1 += 16;
		}
		fwrite(dst, (pitch/8)*32, sizeof(char), ofp);
	}

	// obt@̊J
	delete src;
	delete dst;
}

// 8x16TCYPʂ𐅕@ɕϊ
void ConvertSP2(FILE *ifp, FILE *ofp) {
	char  head[64];
	char *src = NULL;
	char *dst = NULL;
	char *p1, *p2, *p3;
	int   offset;
	int   width;
	int   height;
	int   pitch;
	int   data;
	int   x, y, w, h;

	// wb_[̓ǂݍ
	fread(head, 54, sizeof(char), ifp);
	width   = (head[18] & 255);
	width  |= (head[19] & 255) <<  8;
	width  |= (head[20] & 255) << 16;
	width  |= (head[21] & 255) << 24;
	height  = (head[22] & 255);
	height |= (head[23] & 255) <<  8;
	height |= (head[24] & 255) << 16;
	height |= (head[25] & 255) << 24;
	pitch   = (width+3) & -4;

	// obt@̊m
	src = new char[pitch * 16];
	dst = new char[(pitch/8)*32];

	// sNZϊ
	offset = 54+1024;
	offset += pitch * height;
	for (h = 0; h < height; h += 16) {
		offset -= pitch * 16;
		fseek(ifp, offset, SEEK_SET);
		fread(src, pitch*16, sizeof(char), ifp);
		p1 = src + pitch*16;
		p3 = dst;
		for (w = 0; w < width; w += 8) {
			p2 = p1;
			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p3 += 8;

			for (y = 0; y < 8; y++) {
				p2 -= pitch;
				data = 0;
				for (x = 0; x < 8; x++) {
					data |= ((0x00060500 >> (p2[x] & 3)) & 0x8080) >> x;
				}
				p3[0] = (char)(data     );
				p3[8] = (char)(data >> 8);
				p3++;
			}
			p3 += 8;

			p1 += 8;
		}
		fwrite(dst, (pitch/8)*32, sizeof(char), ofp);
	}

	// obt@̊J
	delete src;
	delete dst;
}

